#include "rclcpp/rclcpp.hpp"
#include "priority_executor/priority_executor.hpp"
#include "priority_executor/priority_memory_strategy.hpp"
#include "casestudy_tools/test_nodes.hpp"
#include "std_msgs/msg/string.hpp"
#include "casestudy_tools/primes_workload.hpp"
#include "casestudy_tools/experiment.hpp"
#include "jsoncpp/json/reader.h"
#include <string>
#include <fstream>
#include <unistd.h>
#include <sys/prctl.h>

std::atomic<bool> should_do_task(true);

void run_one_executor(
    std::function<void(std::atomic<bool>&)> const& exec_fun,
    int const idx) {
    (void)idx;
    
    // set this process to highest priority
    struct sched_param param;
    param.sched_priority = sched_get_priority_max(SCHED_FIFO) - 1;
    int const result = sched_setscheduler(0, SCHED_FIFO, &param);
    if (result != 0) {
        std::cout << "ros_experiment: sched_setscheduler failed: " << result 
                  << ": " << strerror(errno) << std::endl;
    }

    // set process name
    prctl(PR_SET_NAME, "ros_experiment", 0, 0, 0);
    exec_fun(should_do_task);
}

int ros_experiment(
    int const argc,
    char** const argv,
    std::string const& file_name,
    ExperimentConfig config) {

    rclcpp::init(argc, argv);
    
    // we have to create a node to process any passed parameters
    auto const node = rclcpp::Node::make_shared("experiment_node");
    node->declare_parameter("executor_type", "edf");

    std::string executor_type;
    node->get_parameter("executor_type", executor_type);
    std::cout << "using executor type: " << executor_type << std::endl;

    config.executor_type = executor_type;
    config.nodes.push_back(node);

    Experiment experiment(config);
    auto const exec_funs = experiment.getRunFunctions();
    std::cout << "got " << exec_funs.size() << " executors" << std::endl;
    
    std::vector<std::thread> exec_threads;
    int i = 0;
    experiment.resetTimers();
    
    for (auto const& exec_fun : exec_funs) {
        exec_threads.emplace_back(run_one_executor, exec_fun, i++);
    }

    for (auto& t : exec_threads) {
        t.join();
    }

    experiment.writeLogsToFile(file_name, "2023customfile_singlethread");
    return 0;
}

int main(int argc, char* argv[]) {
    std::ifstream input_file("taskset.json");

    if (!input_file.is_open()) {
        std::cerr << "Error opening file!" << std::endl;
        return 1;
    }

    Json::Value root;
    Json::Reader reader;
    bool const parsingSuccessful = reader.parse(input_file, root);

    if (!parsingSuccessful) {
        std::cerr << "Error parsing file!" << std::endl;
        return 1;
    }

    // set ourselves to the second highest priority
    struct sched_param param;
    param.sched_priority = sched_get_priority_max(SCHED_FIFO) - 2;
    int const result = sched_setscheduler(0, SCHED_FIFO, &param);
    if (result != 0) {
        std::cout << "ros_experiment: sched_setscheduler failed: " << result 
                  << ": " << strerror(errno) << std::endl;
    }
    
    // calibrate the dummy load for the current system
    calibrate_dummy_load();
    
    ExperimentConfig config;

    for (uint i = 0; i < root["chain_lengths"].size(); i++) {
        config.chain_lengths.push_back(root["chain_lengths"][i].asUInt());
    }

    for (uint i = 0; i < root["node_ids"].size(); i++) {
        std::vector<uint32_t> node_ids_row;
        for (uint j = 0; j < root["node_ids"][i].size(); j++) {
            node_ids_row.push_back(root["node_ids"][i][j].asUInt());
        }
        config.callback_ids.push_back(std::move(node_ids_row));
    }

    for (uint i = 0; i < root["node_priorities"].size(); i++) {
        config.callback_priorities.push_back(root["node_priorities"][i].asUInt());
    }

    config.num_groups = 0;

    for (uint i = 0; i < root["chain_timer_control"].size(); i++) {
        config.chain_timer_control.push_back(root["chain_timer_control"][i].asInt());
    }

    for (uint i = 0; i < root["node_runtimes"].size(); i++) {
        config.callback_runtimes.push_back(root["node_runtimes"][i].asDouble());
    }

    for (uint i = 0; i < root["chain_periods"].size(); i++) {
        config.chain_periods.push_back(root["chain_periods"][i].asInt());
    }

    for (uint i = 0; i < root["node_executor_assignments"].size(); i++) {
        config.callback_executor_assignments.push_back(root["node_executor_assignments"][i].asUInt() - 1);
    }

    for (uint i = 0; i < root["executor_to_cpu_core"].size(); i++) {
        config.executor_to_cpu_assignments.push_back(root["executor_to_cpu_core"][i].asUInt() - 1);
    }

    config.parallel_mode = false;
    config.cores = 4;

    int const i = (100 - 35) / 5;
    int const ros_budget = 35 + i * 5;
    int const other_task_budget = 65 - i * 5;
    std::cout << "ROS budget: " << ros_budget 
              << " Other task budget: " << other_task_budget << std::endl;

    should_do_task.store(true);
    double const other_task_time = 80 * other_task_budget / 100.0;
    std::thread other_task(do_other_task, other_task_time, 80, std::ref(should_do_task));

    std::string const file_name = "ros_" + std::to_string(ros_budget);
    std::thread ros_task(ros_experiment, argc, argv, file_name, config);
    std::cout << "tasks started" << std::endl;
    ros_task.join();
    other_task.join();
    std::cout << "tasks joined" << std::endl;

    return 0;
}