#include <cmath>
#include <chrono>
#include <string>
#include <fstream>
#include <iostream>
#include <filesystem>
#include <unistd.h>
#include <sys/prctl.h>
#include <rclcpp/rclcpp.hpp>
#include <std_msgs/msg/string.hpp>
#include "casestudy_tools/test_nodes.hpp"
#include "casestudy_tools/experiment.hpp"
#include "casestudy_tools/primes_workload.hpp"
#include "priority_executor/priority_executor.hpp"
#include "priority_executor/priority_memory_strategy.hpp"

std::atomic<bool> should_do_task(true);

// Global performance monitoring configuration
constexpr size_t MONITOR_BUFFER_SIZE = 10000;
constexpr size_t MONITOR_AUTO_DUMP_THRESHOLD = 5000;
const std::string MONITOR_OUTPUT_DIR = "performance_logs/monitored_example";

void configure_executor_monitoring(std::shared_ptr<priority_executor::TimedExecutor> executor, 
                                const std::string& name) {
    executor->setMonitoringOptions(
        MONITOR_BUFFER_SIZE,
        MONITOR_AUTO_DUMP_THRESHOLD,
        MONITOR_OUTPUT_DIR + "/" + name
    );
    
    // Enable monitoring by default
    executor->enableMonitoring(true);
}

void run_one_executor(
    std::function<void(std::atomic<bool>&)> const& exec_fun,
    int const idx) {
    // Set up process scheduling
    struct sched_param param;
    param.sched_priority = sched_get_priority_max(SCHED_FIFO) - 2;
    int const result = sched_setscheduler(0, SCHED_FIFO, &param);
    
    if (result != 0) {
        std::cerr << "ros_experiment: sched_setscheduler failed: " << result 
                  << ": " << strerror(errno) << std::endl;
        std::cerr << "This operation requires root privileges. Please run the program with sufficient permissions." 
                  << std::endl;
#ifndef WIN_DOCKER_IS_BROKEN
        exit(EXIT_FAILURE);
#else
        std::cerr << "Continuing without setting scheduler, fuck Windows" << std::endl;
#endif
    }
    
    prctl(PR_SET_NAME, "ros_experiment", 0, 0, 0);
    exec_fun(should_do_task);
}

int ros_experiment(
    int const argc,
    char** const argv,
    std::string const& file_name,
    ExperimentConfig config) {
    
    rclcpp::init(argc, argv);
    
    auto const node = rclcpp::Node::make_shared("experiment_parameters");
    node->declare_parameter("executor_type", "edf");

    std::string executor_type;
    node->get_parameter("executor_type", executor_type);
    std::cout << "using executor type: " << executor_type << std::endl;

    config.executor_type = executor_type;
    config.parallel_mode = true;
    config.nodes.push_back(node);

    Experiment experiment(config);
    auto const exec_funs = experiment.getRunFunctions();
    std::cout << "got " << exec_funs.size() << " executors" << std::endl;
    
    // Configure monitoring for each executor
    int executor_id = 0;
    std::vector<std::shared_ptr<priority_executor::TimedExecutor>> monitored_executors;
    for (const auto& executor : experiment.getExecutors()) {
        if (auto timed_exec = std::dynamic_pointer_cast<priority_executor::TimedExecutor>(executor.executor)) {
            std::string exec_name = "executor_" + std::to_string(executor_id++);
            configure_executor_monitoring(timed_exec, exec_name);
            monitored_executors.push_back(timed_exec);
        }
    }
    
    std::vector<std::thread> exec_threads;
    int i = 0;
    experiment.resetTimers();
    
    for (auto const& exec_fun : exec_funs) {
        exec_threads.emplace_back(run_one_executor, exec_fun, i++);
    }

    for (auto& t : exec_threads) {
        t.join();
    }
    
    // Write performance reports for each executor
    std::ofstream report_file(MONITOR_OUTPUT_DIR + "/performance_report.txt");
    if (report_file.is_open()) {
        report_file << "===========================\n";
        report_file << " Performance Monitor Report\n";
        report_file << "===========================\n\n";
        
        for (const auto& executor : monitored_executors) {
            report_file << "Executor: " << executor->getExecutorName() 
                      << " (ID: " << executor->getExecutorId() << ")\n";
            report_file << "----------------------------------------\n";
            report_file << executor->getCallbackExecutionReport() << "\n\n";
        }
        
        report_file.close();
        std::cout << "Performance report written to " << MONITOR_OUTPUT_DIR 
                 << "/performance_report.txt" << std::endl;
    } else {
        std::cerr << "Failed to open performance report file" << std::endl;
    }
    
    // Ensure all performance logs are written before experiment logs
    auto& monitor = priority_executor::PerformanceMonitor::getInstance();
    monitor.dumpToFile("final_dump.json");
    
    std::string const outputname = "monitored_example";
    experiment.writeLogsToFile(file_name, outputname);

    return 0;
}

int main(int argc, char* argv[]) {
    // Create the performance logs directory
    std::filesystem::create_directories(MONITOR_OUTPUT_DIR);
    
    // calibrate the dummy load for the current system
    calibrate_dummy_load();

    ExperimentConfig config;
    // Set up two chains of callbacks to demonstrate monitoring
    config.chain_lengths = {2, 2};
    config.callback_ids = {{0, 1}, {2, 3}};
    config.callback_priorities = {1, 0, 3, 2};
    config.chain_timer_control = {0, 1};

    // Set different runtimes to see varying processing times in the logs
    config.callback_runtimes = {10, 20, 15, 25};
    config.chain_periods = {100, 100};
    config.callback_executor_assignments = {};
    config.parallel_mode = true;
    config.cores = 2;

    sanity_check_config(config);

    std::thread ros_task([&]() {
        try {
            ros_experiment(argc, argv, "monitored_example", config);
        } catch (std::exception const& e) {
            std::cerr << "Exception in ros_experiment: " << e.what() << std::endl;
        } catch (...) {
            std::cerr << "Unknown exception in ros_experiment." << std::endl;
        }
    });
    
    std::cout << "tasks started" << std::endl;
    ros_task.join();
    std::cout << "tasks joined" << std::endl;
    
    return 0;
}