#include <chrono>
#include <memory>
#include <string>
#include <thread>

#include "rclcpp/rclcpp.hpp"
#include "std_msgs/msg/string.hpp"

using namespace std::chrono_literals;

constexpr int RUN_DURATION_SECONDS = 5;

// Nodes in the topology:
// - Chain 1: Image2Ground Mapping
//   - Node 1: Camera Node (Sensor Node): Simulates a camera that publishes messages.
//   - Node 2: De-Bayering Node (Processing Node): Processes the camera messages. Listens to Node 1.
//   - Node 3: Radiometric Correction Node (Processing Node): Further processes the messages. Listens to Node 2.
//   - Node 4: Geometric Correction Node (Processing Node): Applies geometric corrections. Listens to Node 3.
//   - Node 5: Image to Ground Mapping Node (Processing Node): Maps the image to ground coordinates. Listens to Node 4.
// - Chain 2: Semantic Scheduler Adaptation
//   - Node 6: Smoke/Fire Classification Node (Decision Node): Listens to Node 3 and makes scheduluer decisions based on the processed data.
// - Chain 3: Flight Control
//   - Node 7: GPS Node (Sensor Node): Simulates a GPS sensor that publishes messages.
//   - Node 8: IMU Node (Sensor Node): Simulates an IMU sensor that publishes messages.
//   - Node 9: Barometric Altimeter Node (Sensor Node): Simulates a barometric altimeter that publishes messages.
//   - Node 10: Sensor Fusion Node (Processing Node): Fuses data from the GPS, IMU, and barometric altimeter. Listens to Nodes 7, 8, and 9.
//   - Node 11: LIDAR Node (Sensor Node): Simulates a LiDAR sensor that publishes messages.
//   - Node 12: Operator Command Node (Sensor Node): Remote control or flight plan issuing node
//   - Node 13: Flight Management Node (Processing Node): Manages flight operations. Listens to Node 10, 11, and 12.
//   - Node 14: Flight Control Node (Processing Node): Controls the flight based on data from Node 13. Listens to Node 13.
// - Chain 4: Data Storage/Streaming
//   - Node 15: Telemetry Node (Processing Node): Collects telemetry data. Listens to Node 10 and 5.
//   - Node 16: Radio Transmitter Node (Processing Node): Transmits data. Listens to Node 15.


// --- Sensor Node ---
class SensorNode : public rclcpp::Node
{
public:
  SensorNode() : Node("sensor_node")
  {
    publisher_ = this->create_publisher<std_msgs::msg::String>("sensor_topic", 10);
    timer_ = this->create_wall_timer(
      100ms, [this]() { publish_message(); });
  }

private:
  void publish_message()
  {
    auto message = std_msgs::msg::String();
    message.data = "Sensor reading";
    RCLCPP_INFO(this->get_logger(), "Publishing: '%s'", message.data.c_str());
    publisher_->publish(message);
  }

  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;
  rclcpp::TimerBase::SharedPtr timer_;
};

// --- Processing Node ---
class ProcessingNode : public rclcpp::Node
{
public:
  ProcessingNode() : Node("processing_node")
  {
    subscription_ = this->create_subscription<std_msgs::msg::String>(
      "sensor_topic", 10,
      [this](std_msgs::msg::String::SharedPtr msg) { process_message(msg); });
    publisher_ = this->create_publisher<std_msgs::msg::String>("processed_topic", 10);
  }

private:
  void process_message(const std_msgs::msg::String::SharedPtr msg)
  {
    RCLCPP_INFO(this->get_logger(), "Processing: '%s'", msg->data.c_str());
    // Simulate processing delay
    std::this_thread::sleep_for(10ms);
    auto new_msg = std_msgs::msg::String();
    new_msg.data = msg->data + " -> processed";
    publisher_->publish(new_msg);
  }

  rclcpp::Subscription<std_msgs::msg::String>::SharedPtr subscription_;
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;
};

// --- Decision Node ---
class DecisionNode : public rclcpp::Node
{
public:
  DecisionNode() : Node("decision_node")
  {
    subscription_ = this->create_subscription<std_msgs::msg::String>(
      "processed_topic", 10,
      [this](std_msgs::msg::String::SharedPtr msg) { decide(msg); });
  }

private:
  void decide(const std_msgs::msg::String::SharedPtr msg)
  {
    RCLCPP_INFO(this->get_logger(), "Deciding based on: '%s'", msg->data.c_str());
    // Future place for semantic detection (e.g., "if smoke detected, boost priority!")
  }

  rclcpp::Subscription<std_msgs::msg::String>::SharedPtr subscription_;
};

// --- Main ---
int main(int argc, char * argv[])
{
  rclcpp::init(argc, argv);

  auto sensor_node = std::make_shared<SensorNode>();
  auto processing_node = std::make_shared<ProcessingNode>();
  auto decision_node = std::make_shared<DecisionNode>();

  rclcpp::executors::MultiThreadedExecutor executor;
  executor.add_node(sensor_node);
  executor.add_node(processing_node);
  executor.add_node(decision_node);

  auto start_time = std::chrono::steady_clock::now();

  while (rclcpp::ok() && (std::chrono::steady_clock::now() - start_time) < std::chrono::seconds(RUN_DURATION_SECONDS))
  {
    executor.spin_some();
    std::this_thread::sleep_for(1ms);
  }

  rclcpp::shutdown();
  return 0;
}
